/*
             Michael Gilleland http://www.merriampark.com/ld.htm

	             converted & hacked from C++ to C
	         2002 Luca Montecchiani <cbm64@inwind.it>
     _________________________________________________________________

                       What is Levenshtein Distance?

   Levenshtein distance (LD) is a measure of the similarity between two
   strings, which we will refer to as the source string (s) and the
   target string (t). The distance is the number of deletions,
   insertions, or substitutions required to transform s into t. For
   example,
     * If s is "test" and t is "test", then LD(s,t) = 0, because no
       transformations are needed. The strings are already identical.
     * If s is "test" and t is "tent", then LD(s,t) = 1, because one
       substitution (change "s" to "n") is sufficient to transform s intot.

   The greater the Levenshtein distance, the more different the strings are.

   Levenshtein distance is named after the Russian scientist Vladimir
   Levenshtein, who devised the algorithm in 1965. If you can't spell or
   pronounce Levenshtein, the metric is also sometimes called edit
   distance.

   The Levenshtein distance algorithm has been used in:
     * Spell checking
     * Speech recognition
     * DNA analysis
     * Plagiarism detection
     _________________________________________________________________

                               The Algorithm
  Steps

   Step Description
   1 Set n to be the length of s.
   Set m to be the length of t.
   If n = 0, return m and exit.
   If m = 0, return n and exit.
   Construct a matrix containing 0..m rows and 0..n columns.
   2 Initialize the first row to 0..n. Initialize the first column to 0..m.
   3 Examine each character of s (i from 1 to n).
   4 Examine each character of t (j from 1 to m).
   5 If s[i] equals t[j], the cost is 0.
   If s[i] doesn't equal t[j], the cost is 1.
   6 Set cell d[i,j] of the matrix equal to the minimum of:
   a. The cell immediately above plus 1: d[i-1,j] + 1.
   b. The cell immediately to the left plus 1: d[i,j-1] + 1.
   c. The cell diagonally above and to the left plus the cost: d[i-1,j-1] + cost.
   7 After the iteration steps (3, 4, 5, 6) are complete, the distance is
   found in cell d[n,m].

                                  Example

   This section shows how the Levenshtein distance is computed when the
   source string is "GUMBO" and the target string is "GAMBOL".

    Steps 1 and 2

       G U M B O
     0 1 2 3 4 5
   G 1
   A 2
   M 3
   B 4
   O 5
   L 6

    Steps 3 to 6 When i = 1

       G U M B O
     0 1 2 3 4 5
   G 1 0
   A 2 1
   M 3 2
   B 4 3
   O 5 4
   L 6 5

    Steps 3 to 6 When i = 2

       G U M B O
     0 1 2 3 4 5
   G 1 0 1
   A 2 1 1
   M 3 2 2
   B 4 3 3
   O 5 4 4
   L 6 5 5

    Steps 3 to 6 When i = 3

       G U M B O
     0 1 2 3 4 5
   G 1 0 1 2
   A 2 1 1 2
   M 3 2 2 1
   B 4 3 3 2
   O 5 4 4 3
   L 6 5 5 4

    Steps 3 to 6 When i = 4

       G U M B O
     0 1 2 3 4 5
   G 1 0 1 2 3
   A 2 1 1 2 3
   M 3 2 2 1 2
   B 4 3 3 2 1
   O 5 4 4 3 2
   L 6 5 5 4 3

    Steps 3 to 6 When i = 5

       G U M B O
     0 1 2 3 4 5
   G 1 0 1 2 3 4
   A 2 1 1 2 3 4
   M 3 2 2 1 2 3
   B 4 3 3 2 1 2
   O 5 4 4 3 2 1
   L 6 5 5 4 3 2

    Step 7

   The distance is in the lower right hand corner of the matrix, i.e. 2.
   This corresponds to our intuitive realization that "GUMBO" can be
   transformed into "GAMBOL" by substituting "A" for "U" and adding "L"
   (one substitution and 1 insertion = 2 changes).
     _________________________________________________________________

     FIXME:
     other LD implementations have separate weight for "addition"
     "change" and "deletion", we can make it configurable.
*/

#include <string.h>
#include <malloc.h>

#include "Simil.h"

#ifdef  min
#undef  min
#endif
#define min(a, b)       ((a)<(b)?(a):(b))

#ifdef  max
#undef  max
#endif
#define max(a, b)       ((a)<(b)?(b):(a))

#define SMALLEST_OF(x,y,z)      ( (x<y) ? min(x,z) : min(y,z) )

//**************************************************
// Get a pointer to the specified cell of the matrix
//**************************************************
int *
GetCellPointer (int *pOrigin, int col, int row, int nCols)
{
	return pOrigin + col + (row * (nCols + 1));
}

//*****************************************************
// Get the contents of the specified cell in the matrix
//*****************************************************
int
GetAt (int *pOrigin, int col, int row, int nCols)
{
	int *pCell;
	pCell = GetCellPointer (pOrigin, col, row, nCols);
	return *pCell;
}

//*******************************************************
// Fill the specified cell in the matrix with the value x
//*******************************************************
void
PutAt (int *pOrigin, int col, int row, int nCols, int x)
{
	int *pCell;
	pCell = GetCellPointer (pOrigin, col, row, nCols);
	*pCell = x;
}

//*****************************
// Compute Levenshtein distance
//*****************************
static int
LD (char *s, char *t)
{
	int *d;			// pointer to matrix
	int n;			// length of s
	int m;			// length of t
	int i;			// iterates through s
	int j;			// iterates through t
	char s_i;		// ith character of s
	char t_j;		// jth character of t
	int cost;		// cost
	int distance;		// distance
	int cell;		// contents of target cell
	int above;		// contents of cell immediately above
	int left;		// contents of cell immediately to left
	int diag;		// contents of cell immediately above and to left
	int sz;			// number of cells in matrix

// Step 1
	n = strlen (s);
	m = strlen (t);
	if (n == 0)
		return m;
	if (m == 0)
		return n;

	sz = (n + 1) * (m + 1) * sizeof (int);
	d = (int *) malloc (sz);

	// Step 2
	for (i = 0; i <= n; i++)
		PutAt (d, i, 0, n, i);
	for (j = 0; j <= m; j++)
		PutAt (d, 0, j, n, j);

	// Step 3
	for (i = 1; i <= n; i++) {
		s_i = s[i - 1];
		// Step 4
		for (j = 1; j <= m; j++) {
			t_j = t[j - 1];
			// Step 5
			if (s_i == t_j)
				cost = 0;
			else
				cost = 1;
			// Step 6
			above = GetAt (d, i - 1, j, n);
			left = GetAt (d, i, j - 1, n);
			diag = GetAt (d, i - 1, j - 1, n);
			cell = SMALLEST_OF (above + 1, left + 1, diag + cost);
			PutAt (d, i, j, n, cell);
		}
	}

	// Step 7
	distance = GetAt (d, n, m, n);
	free (d);

	return (distance);
}

//*****************************************
// Compute Levenshtein distance 0.0 .. 1.0
//*****************************************
float
pre_LD (char *s, char *t)
{
	int distance;		// distance
	float res;		// lenght difference
	float ld_threshold;	// ld difference
	float m_len = (strlen (s) + strlen (t)) / 2.0;	// medium strings lenght

	// preliminary tests
	res = strings_check (s, t);
	if (res != COMPARE)
		return (res);

	// compute !!
	distance = LD (s, t);

	// cool :)
	if (distance == 0)
		return (1.0);

	// distance out of range :(
	if (distance > m_len)
		return (0.0);

	// distance threshold against medium size
	ld_threshold = (float) distance / m_len;

	return (1.0 - ld_threshold);
}
